/**************************************************************************
* CDT30 by Philippe Chavatte - lormedy@free.fr - since 2018, july 25th
* Train controller written for DCCpp and S88 compatibility
* Use a PC serial link with "node.js" engine software, COMx ==> USB to Arduino MEGA with DCCpp software downloaded (@115200kbauds)
* Operates under control of 2 files: CDT30_gui.htmland CDT30_gw.js located in the same directory beside "node_modules" dir
* You can simultaneously control 99 locs and 512 accessories or turnouts on main and programming tracks
* Use 128 step to drive, speed counter is ajustable to your Loc maxi speed.
* Loc number use up to 5 numbers but only the last 4 are used as address.
* Display 2 x 256 sensors of the S88 retro-signalisation bus read thru DCCpp_S88 downloaded to Arduino MEGA
*
* Conduteur de train écrit pour être compatible avec DCCpp et S88 par Philippe Chavatte - lormedy@free.fr - depuis le 25 juillet 2018
* Utilise la liaison série du PC avec le moteur logiciel "node.js", COMx ==> USB vers un Arduino MEGA téléchargé avec le logiciel DCCpp (@115200kbauds)
* Fonctionne sous le controle de 2 fichiers : CDT30_gui.html and CDT30_gw.js placés dans le répertoire qui contient aussi "node_modules"
* Vous pouvez controler simultanément 99 locomotives et 512 accessoires ou aiguillages, sur voie principale et de programmation
* Utilise 128 crans de traction, avec compteur de vitesse réglable à la vitesse maxi de votre loco,
* Numéro de loco sur 5 chiffres mais seulement les 4 derniers sont utilisés comme adresse.
* Affichage du bus de retro-signalisation S88 avec 2 x 256 capteurs lus par DCCpp_S88 téléchargé dans Arduino MEGA
*
* speedmeter and accessories models from Yaasan
* last update by Philippe Chavatte 2019, sept 10th
*
* V1.0 commande DCCpp ou DCC++ - 10/09/2018
* V2.0 lecture et programmation des CV - 10/09/2018
* V2.1 lecture du courant sur voie principale - 15/12/2018
* V3.0 lecture du bus S88 compatible avec DCCpp_S88 - 15/04/2019
***************************************************************************/

var fs = require('fs');
var http = require('http');
var url = require('url');
var sleep = require('system-sleep');
var Serialport = require('serialport');
const Readline = require('@serialport/parser-readline');

var usbserial = "COM3"; // default COM port
var serial_port = null;
var parser = null;
var SerialPort = require('serialport');
var dataReadBack = "";
var wait_answer = 0;

// serial port auto detection
SerialPort.list() // returns promise that resolves to list of ports
	// logs info
	.then(ports => {
		//		console.log(ports);
		ports.forEach(com => {
//			if (com.manufacturer.startsWith('Arduino')) {
			if (com.manufacturer.startsWith('Arduino')||com.manufacturer.startsWith('wch.cn')) {
				usbserial = com.comName;
				serial_port = new Serialport(usbserial, {
					baudRate: 115200,
					// defaults for Arduino serial communication
					dataBits: 8,
					parity: 'none',
					stopBits: 1,
					flowControl: false
				});
				parser = serial_port.pipe(new Readline({ delimiter: '\n' }))
				parser.on('data', readData);

				console.log('System Auto-connect serial port to ' + usbserial);
			}
		});
	});

function readData(data) { // read from serial port
	dataReadBack = data;
	console.log('Data: ' + data);
}

function send2DCC(data2Arduino) {
	var aRetryCounter = 0;

	while (aRetryCounter < 10) {
		console.log('send2DCC: ' + data2Arduino.toString('hex'));

		if (write_sp(data2Arduino) == true) {					// write thru serialPort from PC to Arduino
			return true;
		}

		setTimeout(5);
		aRetryCounter++;
	}
	return false;
}

function write_sp(data) { // write on serial port
		if (serial_port === null) {
		console.log('Serial port is null !!!!!!');
		return false;
	}
	try {
		serial_port.write("<" + data + ">"); // adding DCCpp format
		return true;
	} 
	catch (e) {
		console.log(e);
		return false;
	}
}

fs.readFile('./CDT30_gui.html', function (err, html) {
	if (err) {
		throw err;
	}
	
	var server = http.createServer(
		function (request, response) {

			//console.log(request.url);

			if ( request.method == 'GET' && request.url === "/" ) {
				response.writeHead(200, { 'Content-Type': 'text/html' });
				response.write('<!-- 200 Ok -->\n');
				response.write(html);
				response.end();
			} else if( request.method == 'GET' && request.url.startsWith("/?") ) {
				// wait_answer = 0;
				var url_parts = url.parse(request.url, true);
				//console.log(url_parts.query.CMD);

				//Parse GET_Commands from GUI
				ParseCDTCommand(url_parts.query.CMD);

				if (wait_answer > 0) sleep(wait_answer);	// par tranche de 1ms
				response.writeHead(200, { 'Content-Type': 'text/plain' });
				response.write(dataReadBack);
				response.end();
			} else if (request.url.endsWith(".jpg")) {
				var img = fs.readFileSync('.' + request.url);
				response.writeHead(200, {'Content-Type': 'image/jpg' });
				response.end(img, 'binary');
			} else if (request.url.endsWith(".gif")) {
				var img = fs.readFileSync('.' + request.url);
				response.writeHead(200, {'Content-Type': 'image/gif' });
				response.end(img, 'binary');
			} else if (request.url.endsWith(".png")) {
				var img = fs.readFileSync('.' + request.url);
				response.writeHead(200, {'Content-Type': 'image/png' });
				response.end(img, 'binary');
			} else if (request.url.endsWith(".ico")) {
				var img = fs.readFileSync('.' + request.url);
				response.writeHead(200, {'Content-Type': 'image/ico' });
				response.end(img, 'binary');
			} else if (request.url.endsWith(".js")) {
				var js = fs.readFileSync('.' + request.url);
				response.writeHead(200, { 'Content-Type': 'text/js' });
				response.end(js, 'text/js');
			} else if (request.url.endsWith(".css")) {
				var css = fs.readFileSync('.' + request.url);
				response.writeHead(200, { 'Content-Type': 'text/css' });
				response.end(css, 'text/css');
			} else {
				console.log("error on url: " + request.url);
				response.writeHead(404, { 'Content-Type': 'text/plain' });
				response.write("not found or not implemented");
				response.end();
			}
		}
	).listen(62560);
});

console.log("CDT30: Conducteur de Trains V3.0 - Philippe Chavatte 15/04/2019");
console.log('CDT30 running at http://127.0.0.1:62560/ on PC');
console.log('CDT30 running at http://your-IP:62560/ with Webphone or Tablet');

// global variables
var Byte1 = 0;
var Byte2 = 0;
var Byte128 = 0;
var Byte160 = 0;
var Byte176 = 0;
var Byte222 = 0;
var Byte223 = 0;
var trackSelect = 0;
const MAIN = 1;
const PROG = 0;

function ParseCDTCommand(inStrCommand) {
	var strCommand = "";

	if (inStrCommand == null) {
		return false;
	}

	strCommand = inStrCommand;
	//console.log("Log: " + strCommand);
	strCommand = strCommand.replace("(", ",");
	strCommand = strCommand.replace(")", ",");
	var strParameters = strCommand.split(",");
	/*
	console.log("Func: " + strParameters[0] +
		", " + strParameters[1] +
		", " + strParameters[2] +
		", " + strParameters[3] +
		", " + strParameters[4]);
	*/
	if (strParameters[0] == "setLocoSpeedDir") {			// Vitesse, Direction =================================
		var aAddr = parseInt(strParameters[1]);
		var aSpeed = parseInt(strParameters[2]);
		var aDir = parseInt(strParameters[3]);

		//console.log("A = " + aAddr.toString());
		//console.log(aSpeed.toString());
		//console.log(aDir.toString());

		// <t REGISTER CAB SPEED DIRECTION> SET ENGINE THROTTLES USING 128-STEP SPEED CONTROL

		aSpeed = (aSpeed > 125) ? 126 : aSpeed;		// aSpeed range 0..126
		var reg = (aAddr % 100) + 1;
		//console.log( sprintf ('t%d %d %d %d', reg, aAddr, aSpeed, (aDir & 0x1) ) );
		var param = "";
		param = "t";			// Formatage spécifique du message pour DCC++ ==> throttle
		param += reg;
		param += " ";
		param += aAddr; 		// On met l'adresse de la loco dans le message
		param += " ";
		param += aSpeed;		// Vitesse
		param += " ";
		param += (aDir & 0x1);	// Direction

		//console.log("parameters :", param);

		send2DCC(param);
	}
	else if (strParameters[0] == "setLocoFunction") {		// Fonctions F0..F28 ======================================
		var aAddrFnc = parseInt(strParameters[1]);
		var aFuncNo = parseInt(strParameters[2]);
		var aFuncPw = parseInt(strParameters[3]);

		switch (aFuncNo) {
			case 0: (aFuncPw == 1) ? Byte128 |= 0x90 : Byte128 &= 0x8F; break;
			case 1: (aFuncPw == 1) ? Byte128 |= 0x81 : Byte128 &= 0x9E; break;
			case 2: (aFuncPw == 1) ? Byte128 |= 0x82 : Byte128 &= 0x9D; break;
			case 3: (aFuncPw == 1) ? Byte128 |= 0x84 : Byte128 &= 0x9B; break;
			case 4: (aFuncPw == 1) ? Byte128 |= 0x88 : Byte128 &= 0x97; break;
			case 5: (aFuncPw == 1) ? Byte176 |= 0xB1 : Byte176 &= 0xBE; break;
			case 6: (aFuncPw == 1) ? Byte176 |= 0xB2 : Byte176 &= 0xBD; break;
			case 7: (aFuncPw == 1) ? Byte176 |= 0xB4 : Byte176 &= 0xBB; break;
			case 8: (aFuncPw == 1) ? Byte176 |= 0xB8 : Byte176 &= 0xB7; break;
			case 9: (aFuncPw == 1) ? Byte160 |= 0xA1 : Byte160 &= 0xAE; break;
			case 10: (aFuncPw == 1) ? Byte160 |= 0xA2 : Byte160 &= 0xAD; break;
			case 11: (aFuncPw == 1) ? Byte160 |= 0xA4 : Byte160 &= 0xAB; break;
			case 12: (aFuncPw == 1) ? Byte160 |= 0xA8 : Byte160 &= 0xA7; break;
			case 13: (aFuncPw == 1) ? Byte222 |= 0x01 : Byte222 &= 0xFE; break;
			case 14: (aFuncPw == 1) ? Byte222 |= 0x02 : Byte222 &= 0xFD; break;
			case 15: (aFuncPw == 1) ? Byte222 |= 0x04 : Byte222 &= 0xFB; break;
			case 16: (aFuncPw == 1) ? Byte222 |= 0x08 : Byte222 &= 0xF7; break;
			case 17: (aFuncPw == 1) ? Byte222 |= 0x10 : Byte222 &= 0xEF; break;
			case 18: (aFuncPw == 1) ? Byte222 |= 0x20 : Byte222 &= 0xDF; break;
			case 19: (aFuncPw == 1) ? Byte222 |= 0x40 : Byte222 &= 0xBF; break;
			case 20: (aFuncPw == 1) ? Byte222 |= 0x80 : Byte222 &= 0x7F; break;
			case 21: (aFuncPw == 1) ? Byte223 |= 0x01 : Byte223 &= 0xFE; break;
			case 22: (aFuncPw == 1) ? Byte223 |= 0x02 : Byte223 &= 0xFD; break;
			case 23: (aFuncPw == 1) ? Byte223 |= 0x04 : Byte223 &= 0xFB; break;
			case 24: (aFuncPw == 1) ? Byte223 |= 0x08 : Byte223 &= 0xF7; break;
			case 25: (aFuncPw == 1) ? Byte223 |= 0x10 : Byte223 &= 0xEF; break;
			case 26: (aFuncPw == 1) ? Byte223 |= 0x20 : Byte223 &= 0xDF; break;
			case 27: (aFuncPw == 1) ? Byte223 |= 0x40 : Byte223 &= 0xBF; break;
			case 28: (aFuncPw == 1) ? Byte223 |= 0x80 : Byte223 &= 0x7F; break;
			default: break;
		}

		if (aFuncNo < 5) Byte1 = Byte128;			// selection des bytes à envoyer
		else if (aFuncNo < 9) Byte1 = Byte176;
		else if (aFuncNo < 13) Byte1 = Byte160;
		else if (aFuncNo < 21) {
			Byte1 = 222;
			Byte2 = Byte222;
		}
		else if (aFuncNo < 29) {
			Byte1 = 223;
			Byte2 = Byte223;
		}
		else Byte1 = 0;

		// <f CAB BYTE1 [BYTE2]> OPERATE ENGINE DECODER FUNCTIONS F0-F28

		var param = "";
		param = "f";			// Formatage spécifique du message pour DCC++ ==> fontions
		param += " ";
		param += aAddrFnc; 		// On met l'adresse de la loco dans le message
		param += " ";
		param += Byte1;			// 1er byte
		if ((Byte1 & 0xFE) == 222) {	// deuxieme otet, optionnel
			param += " ";
			param += Byte2;		// 2e byte
		}

		//console.log("param :", param);

		send2DCC(param);
	}	// end of else if (strParameters[0] == "setLocoFunction")
	else if (strParameters[0] == "setTurnout") {			// Aiguillages & accessoires ==============================
		var aAccAddr = parseInt(strParameters[1] & 0x7FF);	// 0..2043
		var aAccSwitch = parseInt(strParameters[2]);

		var address_acc = ((aAccAddr / 4) & 0x1FF); // 0..511
		var sub_address_acc = aAccAddr % 4; 		// 0..3
		var activate_Acc = aAccSwitch & 1;			// 0/1

		// <a ADDRESS SUBADDRESS ACTIVATE> OPERATE STATIONARY ACCESSORY DECODERS

		address_acc &= 0x7F; // 128 x 4 = 512, limite parmi 2044

//		console.log("setTurnout: " + address_acc + "," + sub_address_acc + "," + aAccSwitch);

		var param = "";
		param = "a";				// Formatage spécifique du message pour DCC++ ==> accessoires
		param += " ";
		param += address_acc; 		// On met l'adresse de l'accessoire dans le message
		param += " ";
		param += sub_address_acc;	// sous-adresse
		param += " ";
		param += activate_Acc;		// valeur = 0/1

		//console.log("param :", param);

		send2DCC(param);
	}
	else if (strParameters[0] == "getLocoConfig") {		/// Lire les CVs sur la voie de programmation ==============
		var aAddr = parseInt(strParameters[1]);
		trackSelect = parseInt(strParameters[2]);
		var aCVNo = parseInt(strParameters[3]);
		wait_answer = 1500;								/// 1.5s

		// <R CV CALLBACKNUM CALLBACKSUB> reads a Configuration Variable byte from the Loc decoder on the programming track

		if (trackSelect == PROG) {			// tracks : Prog = 0, Main = 1

		var param = "";
			param = "R";			// Formatage spécifique du message pour DCC++ ==> lire CV
			param += " ";
			param += aCVNo; 		// On met le CV dans le message
			param += " ";
			param += "82";			// 1er Callback : R
			param += " ";
			param += "67";			// 2e Callback  : C

			//console.log("param :", param);

			send2DCC(param);
		}
	}
	else if (strParameters[0] == "setLocoConfig") {		/// Programmer les CVs sur la voie de programmation ========
		var aAddr = parseInt(strParameters[1]);
		var trackSelect = parseInt(strParameters[2]);
		var aCVNo = parseInt(strParameters[3]);
		var aCVValue = parseInt(strParameters[4]);
		wait_answer = 1500;								/// 1.5s

		// <W CV VALUE CALLBACKNUM CALLBACKSUB> WRITE CONFIGURATION VARIABLE BYTE TO ENGINE DECODER ON PROGRAMMING TRACK
		// <w CAB CV VALUE> WRITE CONFIGURATION VARIABLE BYTE TO ENGINE DECODER ON MAIN OPERATIONS TRACK

		var param = "";
		// voie de programmation
		if (trackSelect == PROG) {			// tracks : Prog = 0, Main = 1
			param = "W";			// Formatage spécifique du message pour DCC++ ==> écrire CV
			param += " ";
			param += aCVNo; 		// On met le CV dans le message
			param += " ";
			param += aCVValue;		// On met la valeur du CV dans le message
			param += " ";
			param += "87";			// 1er Callback : W
			param += " ";
			param += "67";			// 2e Callback  : C
		}
		// voie principale
		else {
			param = "w";			// Formatage spécifique du message pour DCC++ ==> écrire CV
			param += " ";
			param += aAddr; 		// On met l'adresse dans le message
			param += " ";
			param += aCVNo; 		// On met le CV dans le message
			param += " ";
			param += aCVValue;		// On met la valeur du CV dans le message
		}

		// console.log("Address : " + aAddr + ", trackSelect : " +  trackSelect + ",  CV n° : " + aCVNo + ", CV to write : " + aCVValue );	// Main
		//console.log("param :", param);

		send2DCC(param);
	}
	else if (strParameters[0] == "setLocoBitConfig") {		/// Programmer les CVs sur la voie de programmation ========
		var aAddr = parseInt(strParameters[1]);
		var trackSelect = parseInt(strParameters[2]);
		var aCVNo = parseInt(strParameters[3]);
		var aCVBitSelect = parseInt(strParameters[4]);
		var aCVBitValue = parseInt(strParameters[5]);
		wait_answer = 1500;									/// 1.5s

		// <B CV BIT VALUE CALLBACKNUM CALLBACKSUB> WRITE CONFIGURATION VARIABLE BIT TO ENGINE DECODER ON PROGRAMMING TRACK
		// <b CAB CV BIT VALUE> WRITE CONFIGURATION VARIABLE BIT TO ENGINE DECODER ON MAIN OPERATIONS TRACK

		var param = "";
		// voie de programmation
		if (trackSelect == PROG) {			// tracks : Prog = 0, Main = 1
			param = "B";			// Formatage spécifique du message pour DCC++ ==> écrire CV bit
			param += " ";
			param += aCVNo; 		// On met le CV dans le message
			param += " ";
			param += aCVBitSelect;	// On met le bit dans le message
			param += " ";
			param += aCVBitValue;	// On met la valeur du bit dans le message
			param += " ";
			param += "87";			// 1er Callback : W
			param += " ";
			param += "98";			// 2e Callback  : b
		}
		// voie principale
		else {
			param = "b";			// Formatage spécifique du message pour DCC++ ==> écrire CV bit
			param += " ";
			param += aAddr; 		// On met l'adresse dans le message
			param += " ";
			param += aCVNo; 		// On met le CV dans le message
			param += " ";
			param += aCVBitSelect;	// On met le bit dans le message
			param += " ";
			param += aCVBitValue;	// On met la valeur du bit dans le message
			param += " ";
		}

		// console.log("Address : " + aAddr + ", trackSelect : " +  trackSelect + ",  CV n° : " + aCVNo + ", CV to write : " + aCVBitValue );	// Main
		//console.log("param :", param);

		send2DCC(param);
	}
	else if (strParameters[0] == "sendCommand") {			// write DCC command ==========================
		var command = strParameters[1];
		wait_answer = (command === "c" || command === "Y") ? 200 : 0;	/// 0.2s
		
		var param = "";
		if (command === "Y") { 
			param += "Y ";
			param += strParameters[2];			/// Formatage spécifique du message pour DCCpp
			param += " ";
			param += strParameters[3];
			wait_answer = 100;
		}
		else {
			param = command;					/// Formatage spécifique du message pour DCCpp
			wait_answer = 200;
		}
		
		send2DCC(param);
	}

}
